package n4e.core.tool;

import java.util.ArrayList;

import n4e.core.database.DatabaseDatasetLoader;
import n4e.core.database.DbAccessRead;
import n4e.core.dbaccess.DbAccessReadExpAttribute;
import n4e.core.dbaccess.DbAccessReadExpCondition;
import n4e.core.dbaccess.DbAccessReadExpRule;
import n4e.core.list.List4SysUser;
import n4e.core.tool.Tool4ExpertSystem;
import n4e.core.tool.Tool4Throwable;

/**
 * The <CODE>Tool4ExpertStochastic</CODE> class provides an expert system with stochastic rules.
 * <P><B>title: SystemCoreExpLib</B></P>
 * <P>description: </P>
 * <P>copyright: (c) 2017 thomas kaffka, born 11/08/1959 in dsseldorf - germany , all rights reserved.</P>
 * @author thomas kaffka , cologne , germany.
 * @version 15.02.2017 15:53:18
 * @since 1.0
 */

public class Tool4ExpertStochastic extends Tool4ExpertSystem {

	private static final String LINE_END = "\r\n";
	
	private int rule_list_len; // IL
	private int attribute_list_len; // RL
	private String expsystem_uniqueid;
	private int attribute_ndx;
	private int count_question;
	private String question;
	private String att_uid;
	private int min_ndx = 0;
	private double curr_min;
	private int result_number;
	private String database;
	private List4SysUser sysUser;
	
	private double[] probability; // P_
	private double[] rule_value; // RV_
	private int[] relevant_symptoms; // EO_
	private double[] min_value; // MI_
	private double[] max_value; // MA_
	private int[] number_symptoms; // IL_
	private boolean[] attribute_active;
	
	private DatabaseDatasetLoader dsl;
	private ArrayList rule_list;
	private ArrayList att_list;
	private Tool4Number t4n;
	private Tool4String t4s;

	/**
	 * The constructor <CODE>Tool4Export</CODE> creates a new instance(object) of the class.
	 * @param sysUser the system user object
	 * @param expsystem_uniqueid the uniqueid of the expert system
	 * @param result_number the number of result items
	 * @param database the database name
	 * @since 1.0
	 */
	public Tool4ExpertStochastic(List4SysUser sysUser, String expsystem_uniqueid, int result_number, String database) {
		super();
		try {
		
			this.sysUser = sysUser;
			this.expsystem_uniqueid = expsystem_uniqueid;
			this.result_number = result_number;
			this.database = database;
			t4n = new Tool4Number();
			t4s = new Tool4String();
			
			dsl = new DatabaseDatasetLoader();
	
			rule_list = dsl.getDataSetRead(sysUser, "ExpRule",
					"ruleid,priorprob,description", null, "idexpcondition", 
					"idexpsystem='" + expsystem_uniqueid + "'",
					"ruleid ASC", database);
			rule_list_len = rule_list.size();
	
			att_list = dsl.getDataSetRead(sysUser, "ExpAttribute",
					"name,question", null, null, 
					"idexpsystem='" + expsystem_uniqueid + "'",
					"name ASC", database);
			attribute_list_len = att_list.size();
	
			probability = new double[rule_list_len];
			rule_value = new double[attribute_list_len];
			relevant_symptoms = new int[rule_list_len];
			min_value = new double[rule_list_len];
			max_value = new double[rule_list_len];
			number_symptoms = new int[rule_list_len];
			attribute_active = new boolean[attribute_list_len];
	
			// find the prior probabilities and rule values.
			for (int i=0;i<rule_list_len;i++) {
				DbAccessReadExpRule rule = (DbAccessReadExpRule) rule_list.get(i); 
				probability[i] = rule.getPriorprob();
				double P = probability[i];
				int len = rule.getMyDbAccessExpCondition().size();
				relevant_symptoms[i] = len;
				number_symptoms[i] = len;
				for (int ndx=0;ndx<len;ndx++) {
					DbAccessRead cond = (DbAccessRead) rule.getMyDbAccessExpCondition().get(ndx);
					DbAccessReadExpCondition condition = new DbAccessReadExpCondition(sysUser, cond.getUniqueid(), "all", "idexpattribute", null);
					if (condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff7") 
							|| condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff6")) { // IF or AND
						int J = getAttNumber(condition.getMyDbAccessReadExpAttribute().getUniqueid());
						ArrayList<Double> prob = getTupleProbabilities(condition.getOperand());
						double PY = prob.get(0);
						double PN = prob.get(1);
						rule_value[J] = rule_value[J] + Math.abs(P * PY / (P * PY + (1 - P) * PN) - P * (1 - PY) / (P * (1 - PY) + (1 - P) * (1 - PN)));
					}
				}
			}
			
			for (int ndx=0;ndx<attribute_list_len;ndx++) {
				attribute_active[ndx] = true;
			}
			
		} catch (Exception ex) {
			System.out.println(Tool4Throwable.getCallStackOut(ex));
		}
	}
	
	/**
     * The method <CODE>findMaxSymptom</CODE> finds the symptom with the max. rule value.
     * @return true: found.
     * @since 1.0
     */
	public boolean findMaxSymptom() {

		// get the index of the maximum rule_value 
		double curr_value = 0;
		attribute_ndx = -1;
		for (int j=0;j<attribute_list_len;j++) {
			if (rule_value[j] > curr_value) {
				attribute_ndx = j;
				curr_value = rule_value[j];
			}
			rule_value[j] = 0;
		}

		if (attribute_ndx == -1) {
			question = "[msg0087]";
			return false;
		}
		
		attribute_active[attribute_ndx] = false;

		count_question++;
		DbAccessReadExpAttribute att = (DbAccessReadExpAttribute) att_list.get(attribute_ndx);
		att_uid = att.getUniqueid();
		question = att.getQuestion() + " (" + count_question + ", " + attribute_ndx + ")" + LINE_END + LINE_END;
		question += "[msg0088]";

		return true;
	}

	/**
     * The method <CODE>updateProbabilities</CODE> update the probabilities.
     * @param answer the answer.
     * @exception Exception if error
     * @since 1.0
     */
	public void updateProbabilities(double answer) throws Exception {

		DbAccessReadExpAttribute att = (DbAccessReadExpAttribute) att_list.get(attribute_ndx);
		putContextList(att.getUniqueid(), Double.toString(answer), -1);
		
		for (int i=0;i<rule_list_len;i++) {
			DbAccessReadExpRule rule = (DbAccessReadExpRule) rule_list.get(i); 
			double P = rule.getPriorprob();
			int len = rule.getMyDbAccessExpCondition().size();
			for (int k=0;k<number_symptoms[i];k++) {
				DbAccessRead cond = (DbAccessRead) rule.getMyDbAccessExpCondition().get(k);
				DbAccessReadExpCondition condition = new DbAccessReadExpCondition(sysUser, cond.getUniqueid(), "all", "idexpattribute", null);
				if (condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff7") 
						|| condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff6")) { // IF or AND
					int J = getAttNumber(condition.getMyDbAccessReadExpAttribute().getUniqueid());
					ArrayList<Double> prob = getTupleProbabilities(condition.getOperand());
					double PY = prob.get(0);
					double PN = prob.get(1);

					if (J != attribute_ndx || relevant_symptoms[i] == 0) {
						continue;
					}
					relevant_symptoms[i] -= 1;
					P = probability[i];
					double PE = P * PY + (1 - P) * PN;
					if (answer > 0) {
						probability[i] = P * (1 + (PY / PE - 1) * answer / 5);
					}
					if (answer <= 0) {
						probability[i] = P * (1 + (PY - (1 - PY) * PE / (1 - PE)) * answer / 5);
					}
					if (probability[i] == Math.floor(probability[i])) {
						relevant_symptoms[i] = 0;
					}
				}
			}
		}
		
		findNewRuleValues();
		
	}

	/**
     * The method <CODE>findNewRuleValues</CODE> calculate new rule values.
     * @exception Exception if error
     * @since 1.0
     */
	private void findNewRuleValues() throws Exception {
		curr_min = 0;
		min_ndx = 0;
		for (int i=0;i<rule_list_len;i++) {
			double P = probability[i];
			double A1 = 1;
			double A2 = 1;
			double A3 = 1;
			double A4 = 1;
			DbAccessReadExpRule rule = (DbAccessReadExpRule) rule_list.get(i); 
			int len = rule.getMyDbAccessExpCondition().size();
			for (int k=0;k<number_symptoms[i];k++) {
				DbAccessRead cond = (DbAccessRead) rule.getMyDbAccessExpCondition().get(k);
				DbAccessReadExpCondition condition = new DbAccessReadExpCondition(sysUser, cond.getUniqueid(), "all", "idexpattribute", null);
				if (condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff7") 
						|| condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff6")) { // IF or AND
					int J = getAttNumber(condition.getMyDbAccessReadExpAttribute().getUniqueid());
					ArrayList<Double> prob = getTupleProbabilities(condition.getOperand());
					double PY = prob.get(0);
					double PN = prob.get(1);

					if (!attribute_active[J] || relevant_symptoms[i] == 0) {
						continue;
					}

					if (PN > PY) {
						PY = 1 - PY;
						PN = 1 - PN;
					}
					rule_value[J] = rule_value[J] + P * PY / (P * PY + (1 - P) * PN) - P * (1 - PY) / (P * (1 - PY) + (1 - P) * (1 - PN));
					A1 = A1 * PY;
					A2 = A2 * PN;
					A3 = A3 * (1 - PY);
					A4 = A4 * (1 - PN);
				}
				max_value[i] = P * A1 / (P * A1 + (1 - P) * A2);
				min_value[i] = P * A3 / (P * A3 + (1 - P) * A4);
				if (max_value[i] < rule.getPriorprob()) {
					relevant_symptoms[i] = 0;
				}
				if (min_value[i] > curr_min) {
					min_ndx = i;
					curr_min = min_value[i];
				}
			}
		}
	}

	/**
     * The method <CODE>searchForBest</CODE> gives the n best answers.
     * @return true: best found.
     * @exception Exception if error
     * @since 1.0
     */
	public boolean searchForBest() throws Exception {
		
		// search the best value.
		for (int i=0;i<rule_list_len;i++) {
			if (min_value[min_ndx] <= max_value[i] && i != min_ndx) {
				return false;
			}
		}

		// find the n best answers.
		if (result_number > attribute_list_len) {
			result_number = attribute_list_len;
		} else if (result_number < 1) {
			result_number = 1;
		}
		
		double[] curr_min_2 = new double[result_number];	
		curr_min_2[0] = curr_min;
		int[] min_ndx_2 = new int[result_number];
		min_ndx_2[0] = min_ndx;
		for (int ndx=1;ndx<result_number;ndx++) {
			for (int i=0;i<rule_list_len;i++) {
				boolean make = true;
				for (int j=0;j<ndx;j++) {
					if (i == min_ndx_2[j]) {
						make = false;
					}
				}
				if (make && min_value[i] > curr_min_2[ndx]) {
					min_ndx_2[ndx] = i;
					curr_min_2[ndx] = min_value[i];
				}
			}
		}

		question = "";
		for (int ndx=0;ndx<result_number;ndx++) {
			DbAccessReadExpRule rule = (DbAccessReadExpRule) rule_list.get(min_ndx_2[ndx]); 
			if (ndx == 0) {
				question += "[msg0089]" + LINE_END;
			} else {
				question += "[msg0090]" + LINE_END;
			}
			question += rule.getDescription() + LINE_END;
			question += "[msg0091]" + LINE_END;
			question += t4n.getNumber2String(probability[min_ndx_2[ndx]] * 100, 4, sysUser.getUserLocale()) + "%" + LINE_END;
			if (result_number > 1) {
				question += "----------------------------------" + LINE_END;
			}
			
			int len2 = rule.getMyDbAccessExpCondition().size();
			for (int k=0;k<len2;k++) {
				DbAccessRead cond = (DbAccessRead) rule.getMyDbAccessExpCondition().get(k);
				DbAccessReadExpCondition condition = new DbAccessReadExpCondition(sysUser, cond.getUniqueid(), "all", "idexpattribute", null);
				if (condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff5")) { // THEN
					putContextList(condition.getMyDbAccessReadExpAttribute().getUniqueid(), 
							rule.getDescription() + " (" + Double.toString(probability[min_ndx_2[ndx]]) + ")", rule.getRuleid());
				}
			}
		}

		return true;
	}

	/**
     * The method <CODE>getQuestion</CODE> gets the current question.
     * @return the question.
     * @since 1.0
     */
	public String getQuestion() {
		return question;
	}
	
	/**
     * The method <CODE>getAttributeUid</CODE> gets the uniqueid of the attribute.
     * @return the uniqueid.
     * @since 1.0
     */
	public String getAttributeUid() {
		return att_uid;
	}
	
	/**
     * The method <CODE>getCountQuestion</CODE> gets the question count.
     * @return the question count.
     * @since 1.0
     */
	public int getCountQuestion() {
		return count_question;
	}
	
	/**
     * The method <CODE>getAttNumber</CODE> gets the number of an attribute.
     * @param att_uid the uniqueid of the attribute.
     * @return the attribute number.
     * @exception Exception if error
     * @since 1.0
     */
	private int getAttNumber(String att_uid) {
		int len = att_list.size();
		for (int ndx=0;ndx<len;ndx++) {
			DbAccessReadExpAttribute att = (DbAccessReadExpAttribute) att_list.get(ndx);
			if (att.getUniqueid().equals(att_uid)) {
				return ndx;
			}
		}
		return -1;
	}

    /**
     * The method <CODE>getTupleProbabilities</CODE> gets the probabilities from the tuple.
     * @param tuple the tuple.
     * @return the attribute number.
     * @exception Exception if error
     * @since 1.0
     */
	private ArrayList<Double> getTupleProbabilities(String tuple) {
		tuple = tuple.trim();
		int ndx1 = tuple.indexOf("(");
		int ndx2 = tuple.indexOf(";");
		int ndx3 = tuple.indexOf(")");
		String spy = tuple.substring(ndx1+1, ndx2).trim();
		String spn = tuple.substring(ndx2+1, ndx3).trim();
		spy = t4s.replaceString(spy, ",", ".");
		spn = t4s.replaceString(spn, ",", ".");
		double py = Double.parseDouble(spy);
		double pn = Double.parseDouble(spn);
		ArrayList<Double> result = new ArrayList<Double>();
		result.add(py);
		result.add(pn);
		return result;
	}
}
