package n4e.core.tool;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.StringTokenizer;

import javax.swing.JDialog;
import javax.swing.JOptionPane;

import n4e.core.database.DatabaseConnection;
import n4e.core.database.DatabaseConnectionSet;
import n4e.core.database.DatabaseDatasetLoader;
import n4e.core.database.DatabaseManager;
import n4e.core.database.DbAccessRead;
import n4e.core.system.GlobalsDialog;
import n4e.core.system.Stack;
import n4e.core.tool.Tool4Throwable;
import n4e.core.dbaccess.DbAccessExpCondition;
import n4e.core.dbaccess.DbAccessExpRule;
import n4e.core.dbaccess.DbAccessReadExpAttribute;
import n4e.core.dbaccess.DbAccessReadExpCondition;
import n4e.core.dbaccess.DbAccessReadExpRule;
import n4e.core.list.Context;
import n4e.core.list.List4SysUser;

/**
 * The <CODE>Tool4ExpertSystem</CODE> class provides the functions of an expert system.
 * <P><B>title: SystemCoreExpLib</B></P>
 * <P>description: </P>
 * <P>copyright: (c) 2013 thomas kaffka, born at 11/08/1959 in dsseldorf - germany , all rights reserved.</P>
 * @author thomas kaffka , cologne , germany.
 * @version 11.02.2013 17:06:46
 * @since 1.0
 */

public class Tool4ExpertSystem {

	protected Hashtable context; // the context list.
	
	/**
	 * The constructor <CODE>Tool4ExpertSystem</CODE> creates a new instance(object) of the class.
	 * @since 1.0
	 */
	public Tool4ExpertSystem() {
		context = new Hashtable();
	}

	/**
	 * The method <CODE>reset</CODE> resets the expert system.
	 * @since 1.0
	 */
	public void reset() {
		context = new Hashtable();
	}
	
	/**
	 * The method <CODE>getContextList</CODE> gets the context list.
	 * @return the context list.
	 * @since 1.0
	 */
	public Hashtable getContextList() {
		return context;
	}
	
	/**
	 * The method <CODE>isValueInSet</CODE> evaluates the context list, 
	 * if the specified attribute has the specified value.
	 * @param att_uid the uniqueid of the attribute.
	 * @param val_uid the uniqueid of the value.  
	 * @return true if available.  
	 * @since 1.0
	 */
	public boolean isValueInSet(String att_uid, String val_uid) {
		ArrayList att_val = getContextListValue(att_uid);
		if (att_val == null) {
			return false;
		}
		int len = att_val.size();
		for (int ndx=0;ndx<len;ndx++) {
			Context con = (Context) att_val.get(ndx);
			if (con.value.equals(val_uid)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * The method <CODE>putContextList</CODE> puts a value into the context list, if not there.
	 * @param key the key.
	 * @param value the value.
	 * @param rule the involved rule number.
	 * @since 1.0
	 */
	public void putContextList(String key, String value, long rule) {
		Context con = new Context(value, rule);
		if (context.containsKey(key)) {
			ArrayList list = (ArrayList) context.get(key);
			int len = list.size();
			for (int ndx=0;ndx<len;ndx++) {
				Context con2 = (Context) list.get(ndx);
				if (con2.value.equals(value)) {
					return;
				}
			}
			list.add(con);
		} else {
			ArrayList list = new ArrayList();
			list.add(con);
			context.put(key, list);
		}
	}
	
	/**
	 * The method <CODE>putListContextList</CODE> puts a list of values into the context list, 
	 * if not there.
	 * @param key the key.
	 * @param value the value.
	 * @param rule the involved rule number.
	 * @since 1.0
	 */
	public void putListContextList(String key, String value, long rule) {
		
		if (value == null || value.trim().length() == 0) {
			putContextList(key, value, rule);
		}
		
		StringTokenizer st = new StringTokenizer(value, "##");
		
		while (st.hasMoreTokens()) {
		
			Context con = new Context(st.nextToken(), rule);
			if (context.containsKey(key)) {
				ArrayList list = (ArrayList) context.get(key);
				int len = list.size();
				for (int ndx=0;ndx<len;ndx++) {
					Context con2 = (Context) list.get(ndx);
					if (con2.value.equals(value)) {
						return;
					}
				}
				list.add(con);
			} else {
				ArrayList list = new ArrayList();
				list.add(con);
				context.put(key, list);
			}
			
		}
	}

	/**
	 * The method <CODE>getContextListValue</CODE> gets the context list value.
	 * @param key the key.
	 * @return the context list value.
	 * @since 1.0
	 */
	public ArrayList getContextListValue(String key) {
		return (ArrayList) context.get(key);
	}

	/**
	 * The method <CODE>getContextListFirstValue</CODE> gets the context list first value.
	 * @param key the key.
	 * @return the context list first value.
	 * @since 1.0
	 */
	public String getContextListFirstValue(String key) {
		ArrayList list = getContextListValue(key);
		if (list == null) {
			return null;
		}
		Context con = (Context) list.get(0);
		return con.value;
	}

	/**
	 * The method <CODE>getContextListAllValues</CODE> gets the context list value list.
	 * @param key the key.
	 * @param bullets true: bullets
	 * @return the context list value list.
	 * @since 1.0
	 */
	public String getContextListAllValues(String key, boolean bullets) {
		ArrayList list = getContextListValue(key);
		if (list == null) {
			return null;
		}
		String result = "";
		int len = list.size();
		for (int ndx=0;ndx<len;ndx++) {
			Context con = (Context) list.get(ndx);
			if (ndx==len-1) {
				result += (bullets?"\\bullet  ":"") + con.value;
			} else {
				result += (bullets?"\\bullet  ":"") + con.value + "\r\n";
			}
		}
		return result;
	}

	/**
	 * The method <CODE>getContextListFirstRule</CODE> gets the context list first rule.
	 * @param key the key.
	 * @return the context list first value.
	 * @since 1.0
	 */
	public long getContextListFirstRule(String key) {
		ArrayList list = (ArrayList) context.get(key);
		Context con = (Context) list.get(0);
		return con.rule;
	}

}

