package n4e.core.tool;

import java.util.ArrayList;
import java.util.StringTokenizer;

import javax.swing.JDialog;
import javax.swing.JOptionPane;

import n4e.core.database.DatabaseDatasetLoader;
import n4e.core.database.DbAccessRead;
import n4e.core.system.Stack;
import n4e.core.tool.Tool4Throwable;
import n4e.core.dbaccess.DbAccessExpRule;
import n4e.core.dbaccess.DbAccessReadExpAttribute;
import n4e.core.dbaccess.DbAccessReadExpCondition;
import n4e.core.dbaccess.DbAccessReadExpRule;
import n4e.core.list.Context;
import n4e.core.list.List4SysUser;

/**
 * The <CODE>Tool4ExpertSystem</CODE> class provides the functions of an expert system with deterministic rules.
 * <P><B>title: SystemCoreExpLib</B></P>
 * <P>description: </P>
 * <P>copyright: (c) 2013 thomas kaffka, born at 11/08/1959 in dsseldorf - germany , all rights reserved.</P>
 * @author thomas kaffka , cologne , germany.
 * @version 11.02.2013 17:06:46
 * @since 1.0
 */

public class Tool4ExpertSystemDeterministic extends Tool4ExpertSystem {

	public static final int ATTRIBUTE = 1;
	public static final int ATTRIBUTE_ANSWER = 2;
	public static final int END = 3;
	
	private String curr_rule_uid;
	
	private Stack goal;
	private List4SysUser sysUser;
	private String database;
	private boolean debug;
	
	private	DatabaseDatasetLoader dsl;
	
	private Tool4String t4s;
	
	/**
	 * The constructor <CODE>Tool4ExpertSystem</CODE> creates a new instance(object) of the class.
	 * @param sysUser the system user object.
	 * @param database the name of the database.
	 * @param debug true: debug on
	 * @since 1.0
	 */
	public Tool4ExpertSystemDeterministic(List4SysUser sysUser, String database, boolean debug) {
		super();
		goal = new Stack();
		this.sysUser = sysUser;
		this.database = database;
		this.debug = debug;
		t4s = new Tool4String();
		try {
			dsl = new DatabaseDatasetLoader();
		} catch (Exception ex) {
			System.out.println(Tool4Throwable.getCallStackOut(ex));
		}
	}

	/**
	 * The method <CODE>reset</CODE> resets the expert system.
	 * @since 1.0
	 */
	public void reset() {
		super.reset();
		goal = new Stack();
	}
	
	/**
	 * The method <CODE>getGoalStack</CODE> gets the goal stack.
	 * @return the goal stack..
	 * @since 1.0
	 */
	public Stack getGoalStack() {
		return goal;
	}
	
	/**
	 * The method <CODE>getCurrRule</CODE> gets the current rule.
	 * @return the rule's uid.
	 * @since 1.0
	 */
	public String getCurrRule() {
		return curr_rule_uid;
	}

	/**
	 * The method <CODE>ruleInterpreter</CODE> performs the knowledge based conclusion.
	 * @return the result
	 * @exception Exception if error
	 * @since 1.0
	 */
	public ArrayList ruleInterpreter() throws Exception {
		
		ArrayList result = new ArrayList();
		
		while (true) {

			String rule_uid = getRuleWithGoal();
		
			if (rule_uid == null) {
				result.add(END);
				return result;
			}
			
			if (isRuleComplete(rule_uid)) {
				
				if (debug) {
					System.out.println("(Tool4ExpertSystem.ruleInterpreter) rule is complete: " + rule_uid);
				}
				
				evaluateRule(rule_uid);
				if (isTopAttributeSingleValue()) {
					goal.pop();
				}
				
			} else {
				
				if (debug) {
					System.out.println("(Tool4ExpertSystem.ruleInterpreter) rule is not complete: " + rule_uid);
				}

				while (true) {
					
					curr_rule_uid = rule_uid;
					String att_uid = getRuleEmptyAtt(rule_uid);
					
					if (att_uid == null) { // rule is complete.
						evaluateRule(rule_uid);
						if (isTopAttributeSingleValue()) {
							goal.pop();
						}
						rule_uid = getRuleWithGoal();
						if (rule_uid == null) {
							result.add(END);
							return result;
						}
						continue;
					}
					
					goal.push(att_uid);
					rule_uid = getRuleWithGoal();
					if (rule_uid == null) {
						goal.pop();
						result.add(ATTRIBUTE);
						result.add(att_uid);
						break;
					}
					
				}
				
				break;
				
			}
		
		}

		return result;
		
	}
	
	/**
	 * The method <CODE>putInitialAttributesOnGoal</CODE> puts the initial attributes 
	 * on the goal stack.
	 * @param curr_expsys_id the uniqueid of the current expert system.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public void putInitialAttributesOnGoal(String curr_expsys_id) throws Exception {
		ArrayList atts = dsl.getDataSetRead(sysUser, "ExpAttribute", 
				null, null, null, "initialattribute=1 AND idexpsystem='" + curr_expsys_id + "'", 
				null, database);
		int len = atts.size();
		if (len == 0) {
			JOptionPane warning = new JOptionPane("please define an initial attribute.", 
					JOptionPane.ERROR_MESSAGE , JOptionPane.CLOSED_OPTION);
			JDialog dialog = warning.createDialog(null, "Error");
			dialog.setVisible(true);
			return;
		}
		for (int ndx=0;ndx<len;ndx++) {
    		DbAccessReadExpAttribute att = (DbAccessReadExpAttribute) atts.get(ndx) ;
    		if (att.errCode == 0) {
    			if (!goal.isObject(att.getUniqueid())) {
    				goal.push(att.getUniqueid());
    			}
    		}
		}
	}
	
	/**
	 * The method <CODE>getRuleWithGoal</CODE> gets the next rule with 
	 * the top of stack goal in the then part.
	 * @return the rule's uniqueid.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public String getRuleWithGoal() throws Exception {
		
		ArrayList rules = dsl.getDataSetRead(sysUser, "ExpRule", 
				null, null, "idexpcondition(condition)", "enabled=1", 
				"ruleid", database);
		
		String top_goal = (String) goal.getTopOfStack();
		if (top_goal == null) {
			return null;
		}

		int len = rules.size();
		for (int ndx=0;ndx<len;ndx++) {
    		DbAccessReadExpRule rule = (DbAccessReadExpRule) rules.get(ndx) ;
    		if (rule.errCode == 0) {
    			ArrayList conds = rule.getMyDbAccessExpCondition();
    			int len2 = conds.size();
				for (int ndx2=0;ndx2<len2;ndx2++) {
					DbAccessRead cond = (DbAccessRead) rule.getMyDbAccessExpCondition().get(ndx2);
					DbAccessReadExpCondition condition = new DbAccessReadExpCondition(sysUser, cond.getUniqueid(), "all", "idexpattribute", null);
		    		if (condition.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff5")) { // THEN.
	    				String att_uid = condition.getMyDbAccessReadExpAttribute().getUniqueid();
	    				if (top_goal.equals(att_uid)) {
	    					return rule.getUniqueid();
	    				}
		    		}
				}
    		} else {
				throw new Exception(rule.errCode + "\r\n" + rule.errMess);
    		}
		}
		
		return null;
	}
	
	/**
	 * The method <CODE>ruleHasElse</CODE> evaluates the conditions of the rule, if ELSE is available.
	 * @param conds the condition list.
	 * @return true, if ELSE is available.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean ruleHasElse(ArrayList conds) throws Exception {
		int len = conds.size();
		for (int ndx=0;ndx<len;ndx++) {
    		DbAccessReadExpCondition con = (DbAccessReadExpCondition) conds.get(ndx);
    		if (con.getCondition().equals("12c5e122:13cf85acfca:-8000")) { /* ELSE */
    			return true;
    		}
		}
		return false;
	}
	
	/**
	 * The method <CODE>isRuleComplete</CODE> evaluates the conditions of the rule.
	 * @param rule_uid the uniqueid of the rule.
	 * @return true, if all conditions has values in the context list.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean isRuleComplete(String rule_uid) throws Exception {
		
		ArrayList conds = dsl.getDataSetRead(sysUser, "ExpCondition", 
				"sort,condition", "idexpattribute(codomain)", null, 
				"idexprule='" + rule_uid + "'", 
				"sort", database);

		int len = conds.size();
		int ndx = 0;
		
		while (ndx<len) {
			
    		DbAccessReadExpCondition con_first = (DbAccessReadExpCondition) conds.get(ndx);
    		
    		if (con_first.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff5")) { // THEN
    			break;
    		}
    		
    		ArrayList<DbAccessReadExpCondition> cond_or_list = new ArrayList<DbAccessReadExpCondition>();
    		if (con_first.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff7") // IF
    				|| con_first.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff6")) { // AND
    			
    			cond_or_list.add(con_first);
    			ndx++;
    			
    			if (ndx<len) {
					con_first = (DbAccessReadExpCondition) conds.get(ndx);
	    			while (con_first.getCondition().equals("-7809eaa8:159267fd635:-7fe7")) { // OR
	        			cond_or_list.add(con_first);
	        			ndx++;
	        			if (ndx>=len) {
	        				break;
	        			}
	    				con_first = (DbAccessReadExpCondition) conds.get(ndx);
	    			}
    			}
    			
    		}
    		
    		int len2 = cond_or_list.size();
    		if (debug) {
    			System.out.println("(Tool4ExpertSystem.isRuleComplete) Condition:");
        		for (int ndx2=0;ndx2<len2;ndx2++) {
            		DbAccessReadExpCondition con = cond_or_list.get(ndx2);
        			if (con.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff7")) {
        				System.out.print(con.getSort() + " IF ");
        			} else if (con.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff6")) {
        				System.out.print(con.getSort() + " AND ");
        			} else if (con.getCondition().equals("-7809eaa8:159267fd635:-7fe7")) {
        				System.out.print(con.getSort() + "   OR ");
        			}
        			System.out.println("");
        		}
        		if (len2==0) {
    				System.out.print("THEN");
        		}
    			System.out.println("(Tool4ExpertSystem.isRuleComplete) ---");
    		}
    		
    		// evaluate all conditions.
    		boolean eval_cond = false;
    		for (int ndx2=0;ndx2<len2;ndx2++) {
    			
        		DbAccessReadExpCondition con = cond_or_list.get(ndx2);

        		String att_uid = con.getMyDbAccessReadExpAttribute().getUniqueid();
				
				// skip function.
				if (con.getMyDbAccessReadExpAttribute().getCodomain().equals("6935473a:13cf7e0dacf:-7fff")) {
					continue;
				}

				if (context.containsKey(att_uid)) {
					eval_cond = evaluateCondition(con.getUniqueid());
					if (eval_cond) {
						break;
					}
				} else {
					return false;
				}

    		}
    		
    		if (!eval_cond) {
				DbAccessExpRule rule = new DbAccessExpRule(sysUser, rule_uid, "enabled", null, null);
				if (rule.errCode != 0) {
					throw new Exception(rule.errCode + "\r\n" + rule.errMess);
				}
				rule.setEnabled(0);
				if (!rule.putToSink(null)) {
					throw new Exception(rule.errCode + "\r\n" + rule.errMess);
				}
				return true;
    		}
    		
		}
		
		return true;
	}
	
	/**
	 * The method <CODE>isTopAttributeSingleValue</CODE> evaluates the top attrinbute on the
	 * goal stack, if it is single value.
	 * @return true, if it is single value.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean isTopAttributeSingleValue() throws Exception {
		String uniqueid = (String) goal.getTopOfStack();
		DbAccessReadExpAttribute att = new DbAccessReadExpAttribute(sysUser, 
				uniqueid, "valuetype,initialattribute,codomain", null, null);
		if (att.errCode != 0) {
			throw new Exception(att.errCode + "\r\n" + att.errMess);
		}
		
		if (att.getValuetype().equals("-72c8c14d:13cbf89ea6b:-7ffa") // single value.
				&& att.getCodomain().equals("-72c8c14d:13cbf89ea6b:-7ffc")) { // set.
			ArrayList values = getContextListValue(att.getUniqueid());
			if (values == null) {
				return false;
			} else {
				return true;
			}
		}
		
		
		if (att.getValuetype().equals("-72c8c14d:13cbf89ea6b:-7ffa")) { // single value.
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * The method <CODE>getRuleEmptyAtt</CODE> gets the rules first attribute 
	 * without a value in the context list.
	 * @param rule_uid the uniqueid of the rule.
	 * @return uniqueid of the attribute.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public String getRuleEmptyAtt(String rule_uid) throws Exception {
		
		ArrayList conds = dsl.getDataSetRead(sysUser, "ExpCondition", 
				"condition", "idexpattribute(name)", null, "idexprule='" + rule_uid + "'", 
				"sort", database);
		
		int len = conds.size();
		for (int ndx=0;ndx<len;ndx++) {
    		DbAccessReadExpCondition con = (DbAccessReadExpCondition) conds.get(ndx);
    		if (con.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff7") /* IF */
    				|| con.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff6")) { /* AND */
				String att_uid = con.getMyDbAccessReadExpAttribute().getUniqueid();
				if (!context.containsKey(att_uid)) {
					return att_uid;
				}
    		}
		}
		
		return null;
	}

	/**
	 * The method <CODE>evaluateBoolean</CODE> evaluates a boolean condition.
	 * @param att_uid the uniqueid of the attribute.
	 * @param operator the operator.
	 * @param soperand the operand as string.
	 * @return true if true.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean evaluateBoolean(String att_uid, String operator, String soperand) throws Exception {
		ArrayList temp_list = (ArrayList) context.get(att_uid);
		Context con = (Context) temp_list.get(0);
		if (con.value == null || con.value.trim().length() == 0) {
			return false;
		}
		boolean value = Boolean.parseBoolean(con.value);
		boolean operand = Boolean.parseBoolean(soperand);
		if (operator.equals("-22c23c05:13cc8f57810:-8000")) { // !=
			return value != operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff2")) { // <
			throw new Exception("boolean no operator <");
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff0")) { // <=
			throw new Exception("boolean no operator <");
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff3")) { // =
			return value == operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff1")) { // >
			throw new Exception("boolean no operator <");
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fef")) { // >=
			throw new Exception("boolean no operator <");
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fee")) { // -->
			return true;
		}
		return false;
	}
	
	/**
	 * The method <CODE>evaluateFloat</CODE> evaluates a float condition.
	 * @param att_uid the uniqueid of the attribute.
	 * @param operator the operator.
	 * @param soperand the operand as string.
	 * @return true if true.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean evaluateFloat(String att_uid, String operator, String soperand) throws Exception {
		ArrayList temp_list = (ArrayList) context.get(att_uid);
		Context con = (Context) temp_list.get(0);
		if (con.value == null || con.value.trim().length() == 0) {
			return false;
		}
		float value = Float.parseFloat(t4s.replaceString(con.value, ",", "."));
		if (soperand.indexOf("..") > -1) {
			StringTokenizer st = new StringTokenizer(soperand, "..");
			float operand1 = Float.parseFloat(t4s.replaceString(st.nextToken(), ",", "."));
			float operand2 = Float.parseFloat(t4s.replaceString(st.nextToken(), ",", "."));
			return operand1 <= value && operand2 >= value;
		}
		float operand = Float.parseFloat(t4s.replaceString(soperand, ",", "."));
		if (operator.equals("-22c23c05:13cc8f57810:-8000")) { // !=
			return value != operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff2")) { // <
			return value < operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff0")) { // <=
			return value <= operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff3")) { // =
			return value == operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff1")) { // >
			return value > operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fef")) { // >=
			return value >= operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fee")) { // -->
			return true;
		}
		return false;
	}
	
	/**
	 * The method <CODE>evaluateInteger</CODE> evaluates an integer condition.
	 * @param att_uid the uniqueid of the attribute.
	 * @param operator the operator.
	 * @param soperand the operand as string.
	 * @return true if true.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean evaluateInteger(String att_uid, String operator, String soperand) throws Exception {
		ArrayList temp_list = (ArrayList) context.get(att_uid);
		Context con = (Context) temp_list.get(0);
		if (con.value == null || con.value.trim().length() == 0) {
			return false;
		}
		int value = Integer.parseInt(con.value);
		if (soperand.indexOf("..") > -1) {
			StringTokenizer st = new StringTokenizer(soperand, "..");
			int operand1 = Integer.parseInt(st.nextToken());
			int operand2 = Integer.parseInt(st.nextToken());
			return operand1 <= value && operand2 >= value;
		}
		int operand = Integer.parseInt(soperand);
		if (operator.equals("-22c23c05:13cc8f57810:-8000")) { // !=
			return value != operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff2")) { // <
			return value < operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff0")) { // <=
			return value <= operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff3")) { // =
			return value == operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff1")) { // >
			return value > operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fef")) { // >=
			return value >= operand;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fee")) { // -->
			return true;
		}
		return false;
	}

	/**
	 * The method <CODE>evaluateSet</CODE> evaluates a set condition.
	 * @param att_uid the uniqueid of the attribute.
	 * @param operator the operator.
	 * @param soperand the operand as string.
	 * @return true if true.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean evaluateSet(String att_uid, String operator, String soperand) throws Exception {
		ArrayList temp_list = (ArrayList) context.get(att_uid);
		Context con = (Context) temp_list.get(0);
		if (con.value == null || con.value.trim().length() == 0) {
			return false;
		}
		String value = con.value;
		String operand = soperand;
		if (operator.equals("-22c23c05:13cc8f57810:-8000")) { // !=
			return !value.equals(operand);
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff3")) { // =
			return value.equals(operand);
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fee")) { // -->
			return true;
		}
		return false;
	}

	/**
	 * The method <CODE>evaluateString</CODE> evaluates a string condition.
	 * @param att_uid the uniqueid of the attribute.
	 * @param operator the operator.
	 * @param soperand the operand as string.
	 * @return true if true.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean evaluateString(String att_uid, String operator, String soperand) throws Exception {
		ArrayList temp_list = (ArrayList) context.get(att_uid);
		Context con = (Context) temp_list.get(0);
		if (con.value == null || con.value.trim().length() == 0) {
			return false;
		}
		String value = con.value;
		String operand = soperand;
		if (operator.equals("-22c23c05:13cc8f57810:-8000")) { // !=
			return !value.equals(operand);
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff2")) { // <
			return value.compareTo(operand) < 0;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff0")) { // <=
			return (value.compareTo(operand) < 0) || value.equals(operand);
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff3")) { // =
			return value.equals(operand);
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7ff1")) { // >
			return value.compareTo(operand) > 0;
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fef")) { // >=
			return (value.compareTo(operand) > 0) || value.equals(operand);
		} else if (operator.equals("-72c8c14d:13cbf89ea6b:-7fee")) { // -->
			return true;
		}
		return false;
	}

	/**
	 * The method <CODE>evaluateCondition</CODE> evaluates a condition.
	 * @param cond_uid the uniqueid of the condition.
	 * @return true if condition is true
	 * @exception Exception if error
	 * @since 1.0
	 */
	public boolean evaluateCondition(String cond_uid) throws Exception {
		
		DbAccessReadExpCondition con = new DbAccessReadExpCondition(sysUser, 
				cond_uid, "operator,operand", "idexpattribute(codomain)", null);
		if (con.errCode != 0) {
			throw new Exception(con.errCode + "\r\n" + con.errMess);
		}
		
		// get the value of the context list.
		String att_uid = con.getMyDbAccessReadExpAttribute().getUniqueid();
		String operator = con.getOperator();
		String operand = con.getOperand();
		String codomain = con.getMyDbAccessReadExpAttribute().getCodomain();

		if (codomain.equals("-72c8c14d:13cbf89ea6b:-7fff")) { // boolean. 
			return evaluateBoolean(att_uid, operator, operand);
		} else if (codomain.equals("-72c8c14d:13cbf89ea6b:-7ffd")) { // float.
			return evaluateFloat(att_uid, operator, operand);
		} else if (codomain.equals("-72c8c14d:13cbf89ea6b:-7ffe")) { // integer.
			return evaluateInteger(att_uid, operator, operand);
		} else if (codomain.equals("-72c8c14d:13cbf89ea6b:-7ffc")) { // set.
			return evaluateSet(att_uid, operator, operand);
		} else if (codomain.equals("-10553fa8:13cc8d95901:-7ffe")) { // string.
			return evaluateString(att_uid, operator, operand);
		} else if (codomain.equals("6935473a:13cf7e0dacf:-7fff")) { // function.
			ArrayList temp_list = (ArrayList) context.get(att_uid);
			if (temp_list == null) {
				putContextList(att_uid, operand, -3);
			}
			return true;
		}
		
		return true;
	}
	
	/**
	 * The method <CODE>evaluateRule</CODE> evaluates the rule.
	 * @param rule_uid the uniqueid of the rule.
	 * @exception Exception if error
	 * @since 1.0
	 */
	public void evaluateRule(String rule_uid) throws Exception {
		
		DbAccessExpRule rule = new DbAccessExpRule(sysUser, rule_uid, "ruleid,enabled", null, null);
		if (rule.errCode != 0) {
			throw new Exception(rule.errCode + "\r\n" + rule.errMess);
		}
		ArrayList conds = dsl.getDataSetRead(sysUser, "ExpCondition", 
				"condition", null, null, "idexprule='" + rule_uid + "'", 
				"sort", database);
		
		// rule is false in one condition and not complete. 
		if (rule.getEnabled() == 0 && !ruleHasElse(conds)) {
			return;
		}
		
		// disable the rule.
		rule.setEnabled(0);
		if (!rule.putToSink(null)) {
			throw new Exception(rule.errCode + "\r\n" + rule.errMess);
		}

		// evaluate the rule.
		boolean is_rule_true = true;
		int then_ndx = -1;
		int else_ndx = -1;
		
		// get the 'then' and 'else' index.
		int len = conds.size();
		for (int ndx=0;ndx<len;ndx++) {
    		DbAccessReadExpCondition con = (DbAccessReadExpCondition) conds.get(ndx);
    		if (con.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff5")) { /* THEN */
    			then_ndx = ndx;
    			continue;
    		}
    		if (con.getCondition().equals("12c5e122:13cf85acfca:-8000")) { /* ELSE */
    			else_ndx = ndx;
    			continue;
    		}
		}

		// evaluate the rule.
		int ndx = 0;
		ArrayList<DbAccessReadExpCondition> cond_or_list = new ArrayList<DbAccessReadExpCondition>();
		
		while (ndx<len) {
			
    		DbAccessReadExpCondition con_first = (DbAccessReadExpCondition) conds.get(ndx);
    		
    		if (con_first.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff5")) { // THEN
    			break;
    		}
    		
    		// get all or conditions.
    		if (con_first.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff7") // IF
    				|| con_first.getCondition().equals("-72c8c14d:13cbf89ea6b:-7ff6")) { // AND
    			
    			cond_or_list.add(con_first);
    			ndx++;

    			if (ndx<len) {
					con_first = (DbAccessReadExpCondition) conds.get(ndx);
	    			while (con_first.getCondition().equals("-7809eaa8:159267fd635:-7fe7")) { // OR
	        			cond_or_list.add(con_first);
	        			ndx++;
	        			if (ndx>=len) {
	        				break;
	        			}
						con_first = (DbAccessReadExpCondition) conds.get(ndx);
	    			}
    			}
    			
    		}

    		// evaluate all or conditions.
    		int len2 = cond_or_list.size();
    		boolean is_cond_true = false;
    		for (int ndx2=0;ndx2<len2;ndx2++) {
    			
        		DbAccessReadExpCondition con = cond_or_list.get(ndx2);
        		
        		if (evaluateCondition(con.getUniqueid())) {
        			is_cond_true = true;
        			break;
        		}

    		}
    		
    		if (!is_cond_true) { 
				is_rule_true = false;
    			break;
    		}
    		
		}

		// all conditions are true, evaluate the 'then' part of the rule.
		if (is_rule_true) {
			if (debug) {
				System.out.println("(Tool4ExpertSystem.evaluateRule) rule is true.");
			}
			DbAccessReadExpCondition con = (DbAccessReadExpCondition) conds.get(then_ndx);
			con = new DbAccessReadExpCondition(sysUser, 
					con.getUniqueid(), "operand", "idexpattribute(name)", null);
			String att_uid = con.getMyDbAccessReadExpAttribute().getUniqueid();
			String operand = con.getOperand();
			putContextList(att_uid, operand, rule.getRuleid());
		} else if (else_ndx > -1) {
			if (debug) {
				System.out.println("(Tool4ExpertSystem.evaluateRule) rule is not true -> else.");
			}
			DbAccessReadExpCondition con = (DbAccessReadExpCondition) conds.get(else_ndx);
			con = new DbAccessReadExpCondition(sysUser, 
					con.getUniqueid(), "operand", "idexpattribute(name)", null);
			String att_uid = con.getMyDbAccessReadExpAttribute().getUniqueid();
			String operand = con.getOperand();
			putContextList(att_uid, operand, rule.getRuleid());
		} else {
			if (debug) {
				System.out.println("(Tool4ExpertSystem.evaluateRule) rule is not true.");
			}
		}
	}

}

