package n4e.core.tool;

import java.util.ArrayList;
import java.util.Random;
import java.util.StringTokenizer;

/**
 * The <CODE>Tool4NeuralNetwork</CODE> class provides the functionality 
 * of a neural network.
 * <P><B>title: SystemCoreExpLib</B></P>
 * <P>description: SystemCoreExpLib</P>
 * <P>copyright: (c) 2017 thomas kaffka, born at 11/08/1959 in dsseldorf - germany , all rights reserved.</P>
 * @author thomas kaffka , cologne , germany.
 * @version 21.02.2017 09:54:30
 * @since 1.0
 */

public class Tool4NeuralNetwork {

	public int errCode;
	
	private double[][] example; // examples
	private double[] factor; // the factor to standardize the x and y values.
	private double[] _x; // input elements.
	private double[] _h; // hidden elements.
	private double[] _y; // output elements.
	private double[] _d; // elements of the desired output.
	private double[][] _w1; // weight matrix layer 1.
	private int w1_i; // index i of the weight matrix layer 1.
	private int w1_j; // index j of the weight matrix layer 1.
	private double[][] _w2; // weight matrix layer 2.
	private int w2_i; // index i of the weight matrix layer 2.
	private int w2_j; // index j of the weight matrix layer 2.
	private int H_len; // size of the hidden layer.
	private double epsilon; // the learn parameter.
	private long learn_step; // counts the learn transit.
	private double max_error_net; // current maximum error net.
	private double min_error_net; // current minimum error net.
	private double max_error; // current maximum error.
	private double min_error; // current minimum error.
	private double average_error; // current average error net.
	private double std_deviation_error; // current standard deviation of the net error.
	private int count_x; // count of the x-coordinates.
	private int count_y; // count of the y-coordinates.
	private boolean std_deviation;

	/**
	 * The constructor <CODE>Tool4NeuralNetwork</CODE> creates a new instance(object) of the class.
	 * @param examples the example data.
	 * @param x_layer the number of the elements in the x layer.
	 * @param y_layer the number of the elements in the y layer.
	 * @param hidden size of the hidden layer.
	 * @param epsilon the learn parameter epsilon.
	 * @param std_deviation thue = calculate the standard deviation.
	 * @since 1.0
	 */
	public Tool4NeuralNetwork(ArrayList<ArrayList<Number>> examples, long x_layer, long y_layer, long hidden, double epsilon, boolean std_deviation) {
		
		errCode = 0;
		
		this.std_deviation = std_deviation;
		
		count_x = (int) x_layer;
		count_y = (int) y_layer;
		
		// get the example data.
		if (examples == null) {
			return;
		}
		
		setExamples(examples);
		
		// initialize the neural network.
		_x = new double[count_x + 1];
		_y = new double[count_y];
		_d = new double[count_y];

		this.epsilon = epsilon;
		H_len = (int) hidden;
		
		learn_step = 0;
		_x[count_x] = 1;
		_h = new double[H_len + 1];
		_h[H_len] = 1;
		w1_i = count_x + 1;
		w1_j = H_len + 1;
		_w1 = new double[w1_i][w1_j];
		w2_i = H_len + 1;
		w2_j = count_y; 
		_w2 = new double[w2_i][w2_j];
		
		// initialize weights with random numbers.
		Random rand = new Random();
		
		for (int i = 0; i < H_len + 1; i++) {
			for (int j = 0; j < count_x + 1; j++) {
				_w1[j][i] = rand.nextDouble();
			}
		}
		
		for (int i = 0; i < H_len + 1; i++) {
			for (int j = 0; j < count_y; j++) {
				_w2[i][j] = rand.nextDouble();
			}
		}
	}

	public long getLearnStep() {
		return learn_step;
	}

	public void setLearnStep(long learn_step) {
		this.learn_step = learn_step;
	}

	public double getMaxError() {
		return max_error;
	}
	
	public double getMinError() {
		return min_error;
	}

	public double getAverageError() {
		return average_error;
	}

	public double getStdDeviation() {
		return std_deviation_error;
	}

	public double[][] getW1() {
		return _w1;
	}

	public void setW1(double[][] input) {
		_w1 = input;
	}

	public double[][] getW2() {
		return _w2;
	}

	public void setW2(double[][] input) {
		_w2 = input;
	}

	public double[] getFactor() {
		return factor;
	}

	public void setExamples(ArrayList<ArrayList<Number>> examples) {

		// prepare the factor processing.
		double[] max_value = new double[count_x + count_y];
		int len = max_value.length;
		for (int ndx=0;ndx<len;ndx++) {
			max_value[ndx] = Double.MIN_VALUE;
		}

		try {
			
			len = examples.size();
			example = new double[len][count_x + count_y];
			for (int ndx=0;ndx<len;ndx++) {
				ArrayList<Number> line = examples.get(ndx);
				for (int ndx2=0;ndx2<count_x+count_y;ndx2++) {
					example[ndx][ndx2] = (double) line.get(ndx2);
					if (example[ndx][ndx2] > max_value[ndx2]) {
						max_value[ndx2] = example[ndx][ndx2];
					}
				}
			}

		} catch (Exception ex) {
			String mess = Tool4Throwable.getCallStackOut(ex);
			if (mess.indexOf("Index:") > -1) {
				errCode = -1; // example don't match.
				return;
			}
		}

		// calculate the standardization factors.
		len = max_value.length;
		factor = new double[len];
		for (int ndx=0;ndx<len;ndx++) {
			if (max_value[ndx] > 1.0) {
				factor[ndx] = 1.0 / (max_value[ndx] + 1.0);
			} else {
				factor[ndx] = 1.0;
			}
		}
	}
	
	/**
	 * The method <CODE>setValues</CODE> set the values for executing the neural network.
	 * @param H the hidden layer. 
	 * @param count_x the number of the x coordinates.
	 * @param count_y the number of the y coordinates.
	 * @param factor the data factor.
	 * @param epsilon the learn parameter epsilon.
	 * @since 1.0
	 */
	public void setValues(int H, int count_x, int count_y, double[] factor, double epsilon) {
		
		int len = factor.length;
		this.factor = new double[len];
		for (int ndx=0;ndx<len;ndx++) {
			this.factor[ndx] = factor[ndx];
		}
		
		this.epsilon = epsilon;
		
		H_len = H;
		this.count_x = count_x;
		this.count_y = count_y;
		
		_x = new double[count_x + 1];
		_y = new double[count_y];
		_d = new double[count_y];
		
		_x[count_x] = 1;
		
		_h = new double[H_len + 1];
		_h[H_len] = 1;
		
		w1_i = count_x + 1;
		w1_j = H_len + 1;
		_w1 = new double[w1_i][w1_j];
		
		w2_i = H_len + 1;
		w2_j = count_y; 
		_w2 = new double[w2_i][w2_j]; 
	}
	
	/**
	 * The method <CODE>backPropagate</CODE> the back propagation.
	 * @since 1.0
	 */
	private void backPropagate() {
		
		// gradient descend.
		double[] Err_H = new double[H_len + 1]; // error of the hidden elements.

		for (int k = 0; k < H_len + 1; k++) {
			Err_H[k] = 0.0;
		}
		max_error_net = Double.MIN_VALUE;
		min_error_net = Double.MAX_VALUE;
		
		for (int j = 0; j < count_y; j++) { // adjustment of the w2 weights, to calculate 
									        // the error in the elements H[k], Err_H[k].
			double delta = Math.abs(_d[j] - _y[j]);
			if (delta > max_error_net) max_error_net = delta; // error in the y layer.
			if (delta < min_error_net) min_error_net = delta; // error in the y layer.
			
			delta = (_d[j] - _y[j]) * _y[j] * (1.0 - _y[j]);
			
			for (int k = 0; k < H_len + 1; k++) {
				Err_H[k] += delta * _w2[k][j];
				_w2[k][j] += epsilon * delta * _h[k]; // deviation.
			}
			
		}
		
		for (int k = 0; k < H_len; k++) { // adjustment W1
			double delta = Err_H[k] * _h[k] * (1.0 - _h[k]);
			for (int i = 0; i < count_x + 1; i++) {
				_w1[i][k] += epsilon * delta * _x[i]; // deviation.
			}
		}
	}

	/**
	 * The method <CODE>t</CODE> the transfer function (sigmoid).
	 * @param a the action of a node.
	 * @return the activity of the node.
	 * @since 1.0
	 */
	private double t(double a) {
		return (double) (1.0 / (1.0 + Math.exp(-a)));
	}

	/**
	 * The method <CODE>executeFunction</CODE> executes the function f(x).
	 * @param x the arguments of the function.
	 * @return the function values.
	 * @since 1.0
	 */
	public double[] executeFunction(double[] x) {
		
		try {
		
			int lenx = x.length;
			for (int ndx=0;ndx<lenx;ndx++) {
				_x[ndx] = x[ndx] * factor[ndx]; 
			}
			
			execute();
			
			double[] result = new double[_y.length];
			int len = _y.length;
			for (int ndx=0;ndx<len;ndx++) {
				result[ndx] = _y[ndx] / factor[lenx + ndx];
			}
		
			return result;

		} catch (Exception ex) {
			return null;
		}
		
	}
	
	/**
	 * The method <CODE>execute</CODE> executes the neural network to get its result.
	 * @since 1.0
	 */
	private void execute() {
		
		for (int k = 0; k < H_len; k++) {
			_h[k] = 0.0;
		}
		
		for (int j = 0; j < count_y; j++) {
			_y[j] = 0.0; 
		}
		
		for (int k = 0; k < H_len; k++) { // for each element of the hidden layer.
			double a = 0.0; // compute its activation.
			for (int i = 0; i < count_x + 1; i++) {
				a += (_w1[i][k] * _x[i]); // activation.
			}
			_h[k] = t(a); // apply the transfer function upon the activation.
		}
		
		for (int j = 0; j < count_y; j++) { // for each element of the y layer.
			double a = 0.0; // compute its activation.
			for (int k = 0; k < H_len + 1; k++) {
				a += (_w2[k][j] * _h[k]); // activation.
			}
			_y[j] = t(a); // apply the transfer function upon the activation.
		}
	}

	/**
	 * The method <CODE>learning</CODE> the learning procedure.
	 * @since 1.0
	 */
	public void learning() {
		
		learn_step++;
		
		// learning of the net.
		max_error = Double.MIN_VALUE;
		min_error = Double.MAX_VALUE;
		average_error = 0.0;
		
		double summ_XQ = 0.0; // coefficient for the standard deviation.
		double summ_Xi = 0.0; // coefficient for the standard deviation.
		
		int len = example.length;
		for (int p = 0; p < len; p++) {
			// transfer the examples to the net.
			for (int i=0;i<count_x;i++) {
				_x[i] = example[p][i] * factor[i];
			}
			for (int j=0;j<count_y;j++) {
				_d[j] = example[p][count_x + j] * factor[count_x + j];
			}

			// learn.
			execute();
			backPropagate();
			
			// prepare the error computing.
			if (max_error_net > max_error) max_error = max_error_net;
			if (min_error_net < min_error) min_error = min_error_net;
			
			average_error += max_error_net;
			if (std_deviation) {
				summ_XQ += (double) Math.pow(max_error_net, 2);
				summ_Xi += max_error_net;
			}
		}
		
		double N = len;
		average_error = average_error / N;
		if (std_deviation) {
			double variance = (1 / (N - 1)) * (summ_XQ - (1 / N * Math.pow(summ_Xi, 2)));
			std_deviation_error = Math.sqrt(variance);
		}
	}
}

